/**
 * vim: set ts=4 :
 * =============================================================================
 * Builtin Votes BaseVote Class
 * Copyright (C) 2011 Ross Bemrose (Powerlord).  All rights reserved.
 * =============================================================================
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, version 3.0, as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * As a special exception, AlliedModders LLC gives you permission to link the
 * code of this program (as well as its derivative works) to "Half-Life 2," the
 * "Source Engine," the "SourcePawn JIT," and any Game MODs that run on software
 * by the Valve Corporation.  You must obey the GNU General Public License in
 * all respects for all other code used.  Additionally, AlliedModders LLC grants
 * this exception to all derivative works.  AlliedModders LLC defines further
 * exceptions, found in LICENSE.txt (as of this writing, version JULY-31-2007),
 * or <http://www.sourcemod.net/license.php>.
 *
 * Version: $Id$
 */

#ifndef _INCLUDE_BUILTINVOTES_BASEVOTE_H_
#define _INCLUDE_BUILTINVOTES_BASEVOTE_H_

//#include <IShareSys.h>
//#include <IHandleSys.h>
//#include <IGameHelpers.h>
//#include <IForwardSys.h>
#include "smsdk_ext.h"
#include "sh_vector.h"
#include "icvar.h"
#include "convar.h"
#include "igameevents.h"

using namespace SourceHook;
using namespace SourceMod;

#define SERVER_CLIENT_INDEX 99 // Defined by TF2
#define MAX_VOTE_OPTIONS 5 // Defined by TF2

#define TEAMS_ALL 255

#define VOTE_YES_INDEX 0
#define VOTE_NO_INDEX 1
#define VOTE_YES 1
#define VOTE_NO 0

// Server vote to change maps. TF2 only.
#define TRANSLATION_TF2_VOTE_MAP_START					"#TF_vote_nextlevel_choices"
#define TRANSLATION_TF2_VOTE_MAP_EXTEND_PASSED			"#TF_vote_passed_nextlevel_extend"

// User vote to kick user.
#define TRANSLATION_TF2_VOTE_KICK_IDLE_START			"#TF_vote_kick_player_idle"
#define TRANSLATION_TF2_VOTE_KICK_SCAMMING_START		"#TF_vote_kick_player_scamming"
#define TRANSLATION_TF2_VOTE_KICK_CHEATING_START		"#TF_vote_kick_player_cheating"
#define TRANSLATION_TF2_VOTE_KICK_START					"#TF_vote_kick_player_other"
#define TRANSLATION_TF2_VOTE_KICK_PASSED				"#TF_vote_passed_kick_player"
#define TRANSLATION_L4D_VOTE_KICK						"#L4D_vote_kick_player"
#define TRANSLATION_L4D_VOTE_KICK_PASSED				"#L4D_vote_passed_kick_player"

// User vote to restart map.
#define TRANSLATION_TF2_VOTE_RESTART_START				"#TF_vote_restart_game"
#define TRANSLATION_TF2_VOTE_RESTART_PASSED				"#TF_vote_passed_restart_game"
#define TRANSLATION_L4D_VOTE_RESTART_START				"#L4D_vote_restart_game"
#define TRANSLATION_L4D_VOTE_RESTART_PASSED				"#L4D_vote_passed_restart_game"

// User vote to change maps.
#define TRANSLATION_TF2_VOTE_CHANGELEVEL_START			"#TF_vote_changelevel"
#define TRANSLATION_TF2_VOTE_CHANGELEVEL_PASSED			"#TF_vote_passed_changelevel"
#define TRANSLATION_L4D_VOTE_CHANGECAMPAIGN_START		"#L4D_vote_mission_change"
#define TRANSLATION_L4D_VOTE_CHANGECAMPAIGN_PASSED		"#L4D_vote_passed_mission_change"
#define TRANSLATION_L4D_VOTE_CHANGELEVEL_START			"#L4D_vote_chapter_change"
#define TRANSLATION_L4D_VOTE_CHANGELEVEL_PASSED			"#L4D_vote_passed_chapter_change"

// User vote to change next level. TF2 only.
#define TRANSLATION_TF2_VOTE_NEXTLEVEL_SINGLE_START		"#TF_vote_nextlevel"
#define TRANSLATION_TF2_VOTE_NEXTLEVEL_MULTIPLE_START	"#TF_vote_nextlevel_choices" // Started by server
#define TRANSLATION_TF2_VOTE_NEXTLEVEL_EXTEND_PASSED	"#TF_vote_passed_nextlevel_extend"
#define TRANSLATION_TF2_VOTE_NEXTLEVEL_PASSED			"#TF_vote_passed_nextlevel"

// User vote to return to lobby. L4D/L4D2 only.
#define TRANSLATION_L4D_VOTE_RETURNTOLOBBY_START		"#L4D_vote_return_to_lobby"
#define TRANSLATION_L4D_VOTE_RETURNTOLOBBY_PASSED		"#L4D_vote_passed_return_to_lobby"

// User vote to change difficulty. L4D/L4D2 only.
#define TRANSLATION_L4D_VOTE_CHANGEDIFFICULTY_START		"#L4D_vote_change_difficulty"
#define TRANSLATION_L4D_VOTE_CHANGEDIFFICULTY_PASSED	"#L4D_vote_passed_change_difficulty"

// User vote to scramble teams.  Can be immediate or end of round. TF2 only.
#define TRANSLATION_TF2_VOTE_SCRAMBLE_IMMEDIATE_START	"#TF_vote_scramble_teams"
#define TRANSLATION_TF2_VOTE_SCRAMBLE_ROUNDEND_START	"#TF_vote_should_scramble_round"
#define TRANSLATION_TF2_VOTE_SCRAMBLE_PASSED 			"#TF_vote_passed_scramble_teams"

// While not a vote string, it works just as well.
#define TRANSLATION_TF2_VOTE_CUSTOM						"#TF_playerid_noteam"
#define TRANSLATION_L4D_VOTE_CUSTOM						"#L4D_TargetID_Player"

/**
 * @brief Reasons a vote can end.
 */
enum BuiltinVoteEndReason
{
	BuiltinVoteEnd_Selected = 0,			/**< Menu item was selected */
	BuiltinVoteEnd_Done = -1,			/**< Voting finished */
	BuiltinVoteEnd_Cancelled = -2,		/**< Voting was cancelled */
};

/**
 * @brief Reasons a vote can be cancelled.
 *
 * These numbers may seem random, but they are used in TF2 as given
 * They are are passed to the client VoteFail handler via UserMessage
 */
enum BuiltinVoteFailReason
{
	BuiltinVoteFail_Generic = 0,		/**< Vote was generically cancelled. */
	BuiltinVoteFail_Loses = 3,			/**< No votes outnumbered Yes votes */
	BuiltinVoteFail_NotEnoughVotes = 4,	/**< Vote did not receive enough votes. */
};

/**
 * @brief Vote types. These are mapped to translation strings and pass strings by VoteStart and VotePass handlers
 */
enum BuiltinVoteType
{
	BuiltinVoteType_Kick = 0,			/**< Yes/No, argument is player userid */
	BuiltinVoteType_KickIdle,			/**< TF2: Yes/No, argument is player userid */
	BuiltinVoteType_KickScamming,		/**< TF2: Yes/No, argument is player userid */
	BuiltinVoteType_KickCheating,		/**< TF2: Yes/No, argument is player userid */
	BuiltinVoteType_Restart,			/**< Yes/No, argument ignored */
	BuiltinVoteType_ChangeLevel,		/**< Yes/No, argument is map name in TF2 or level number in L4D/L4D2 */
	BuiltinVoteType_ChangeCampaign,		/**< L4D/L4D2: Yes/No, argument is campaign name */
	BuiltinVoteType_ReturnToLobby,		/**< L4D/L4D2: Yes/No, argument ignored */
	BuiltinVoteType_ChangeDifficulty,	/**< L4D/L4D2: Yes/No, argument is difficulty level */
	BuiltinVoteType_NextLevel,			/**< TF2: Yes/No, argument is map name */
	BuiltinVoteType_NextLevelChoices,	/**< TF2: Multiple-choice, argument ignored */
	BuiltinVoteType_ScrambleImmediate,	/**< TF2: Yes/No, argument ignored */
	BuiltinVoteType_ScrambleRoundEnd,	/**< TF2: Yes/No, argument ignored */
	BuiltinVoteType_Custom				/**< Yes/No or Multiple-choice (TF2 only), argument is vote text. */
};

enum BuiltinVoteAction
{
	BuiltinVoteAction_Start = (1<<0),		/**< A menu has been started (nothing passed). Fires once per client. */
	BuiltinVoteAction_Select = (1<<1),		/**< An item was selected (param1=client, param2=item) */
	BuiltinVoteAction_End = (1<<2),			/**< A vote's display/selection cycle is complete (nothing passed). */
	BuiltinVoteAction_VoteEnd = (1<<3),		/**< A vote sequence has ended (param1=chosen item) */
	BuiltinVoteAction_VoteStart = (1<<4), 	/**< A vote sequence has started. Fires once. */
	BuiltinVoteAction_Cancel = (1<<5),		/**< A vote sequence has been cancelled (nothing passed) */
};

// Blatant copy of IMenuManager's menu_vote_result_t
// It's a new struct rather than including all the menu functions we don't need.
struct builtinvote_result_t
{
	unsigned int num_clients;		/**< Number of clients the vote was displayed to */
	unsigned int num_votes;			/**< Number of votes received */
	struct client_vote_t
	{
		int client;					/**< Client index */
		int item;					/**< Item # (or -1 for none) */
	} *client_list;					/**< Array of size num_clients */
	unsigned int num_items;			/**< Number of items voted for */
	struct item_vote_t
	{
		unsigned int item;			/**< Item index */
		unsigned int count;			/**< Number of votes */
	} *item_list;					/**< Array of size num_items, sorted by count, 
											descending */
};

class BaseVote
{
public:
	BaseVote(IPluginFunction *pHandler, IdentityToken_t *pOwner=NULL, BuiltinVoteType voteType=BuiltinVoteType_Custom, bool bYesNo=true);
	bool AppendItem(const char *info);
	bool InsertItem(unsigned int position, const char *info);
	bool RemoveItem(unsigned int position);
	void RemoveAllItems();
	const char *GetItemInfo(unsigned int position);
	unsigned int GetItemCount();
	BuiltinVoteType GetVoteType();
	const char *GetVoteArgument();
	void SetVoteArgument(const char *voteArgument);
	void SetResultHandler(IPluginFunction *pVoteResults);
	void Destroy(bool releaseHandle);
	void Cancel();
	bool IsYesNoVote();
	int GetVoteInitiator();
	Handle_t GetHandle();
	void SetTeam(byte team);
	byte GetTeam();
public: // Different for each subclass
	virtual unsigned int GetApproxMemUsage() =0;
	virtual bool SetVoteType(BuiltinVoteType voteType) =0;
	virtual void DisplayVotePass(const char *param1="") =0;
	virtual void DisplayVoteFail(BuiltinVoteFailReason voteReason) =0;
	virtual void CastVote(edict_t *pEntity, const CCommand &args) =0;
public: // Internal implementation of various actions inside the vote
	virtual void ShowVote(cell_t clients[], unsigned int playerCount) =0;
	virtual void ShowVoteCast(int entityId, int item) =0;
	virtual void DisplayAllVotes(unsigned int totalClients, unsigned int yesVotes, unsigned int noVotes);
	virtual void SendOptions();
public: // Old IBuiltinVoteHandler stuff, callbacks mainly.  Can be used by internal code as well as natives
	void OnBuiltinVoteStart();
	void OnBuiltinVoteSelect(int client, unsigned int item);
	void OnBuiltinVoteEnd(BuiltinVoteEndReason reason);
	void OnBuiltinVoteDestroy();
	void OnBuiltinVoteVoteStart();
	void OnBuiltinVoteResults(const builtinvote_result_t *results);
	void OnBuiltinVoteCancel(BuiltinVoteFailReason reason);
	cell_t DoAction(BuiltinVoteAction action, cell_t param1, cell_t param2, cell_t def_res=0);
protected:
	void InternalDelete();
	unsigned int GetAllPlayers(cell_t clients[]);
protected: // Variables
	IdentityToken_t *m_pOwner;
	BuiltinVoteType m_voteType;
	int m_Flags;
	int m_voteInitiator;
	Handle_t m_hHandle;
	bool m_bYesNo;
	CVector<const char *> m_pVoteOptions;
	bool m_bCancelling;
	bool m_bDeleting;
	bool m_bShouldDelete;
	bool m_bWillFreeHandle;
	cell_t m_clients[256];
	unsigned int m_playerCount;
	const char *m_voteArgument;
	IPluginFunction *m_pVoteResults;
	IPluginFunction *m_pHandler;
	byte m_Team;
};

class L4DVote : public BaseVote
{
public:
	L4DVote(IPluginFunction *m_pHandler, IdentityToken_t *pOwner=NULL, BuiltinVoteType voteType=BuiltinVoteType_Custom);
public: // Different for each subclass
	unsigned int GetApproxMemUsage();
	bool SetVoteType(BuiltinVoteType voteType);
	void DisplayVotePass(const char *param1="");
	void DisplayVoteFail(BuiltinVoteFailReason voteReason);
	void CastVote(edict_t *pEntity, const CCommand &args);
protected: // Internal implementation of various actions inside the vote
	void ShowVote(cell_t clients[], unsigned int playerCount, const char *voteArgument);
	void ShowVoteCast(int entityId, int item);
	void DisplayAllVotes(unsigned int totalClients, unsigned int yesVotes, unsigned int noVotes);
	//bool SendOptions();
private:
	void VoteEnded();
};

class L4D2Vote : public BaseVote
{
public:
	L4D2Vote(IPluginFunction *m_pHandler, IdentityToken_t *pOwner=NULL, BuiltinVoteType voteType=BuiltinVoteType_Custom);
public: // Different for each subclass
	unsigned int GetApproxMemUsage();
	bool SetVoteType(BuiltinVoteType voteType);
	void DisplayVotePass(const char *param1="");
	void DisplayVoteFail(BuiltinVoteFailReason voteReason);
	void CastVote(edict_t *pEntity, const CCommand &args);
protected: // Internal implementation of various actions inside the vote
	void ShowVote(cell_t clients[], unsigned int playerCount, const char *voteArgument);
	void ShowVoteCast(int entityId, int item);
	void DisplayAllVotes(unsigned int totalClients, unsigned int yesVotes, unsigned int noVotes);
	//bool SendOptions();
private:
	void VoteEnded();
};

class TF2Vote : public BaseVote
{
public:
	TF2Vote(IPluginFunction *m_pHandler, IdentityToken_t *pOwner=NULL, BuiltinVoteType voteType=BuiltinVoteType_Custom, bool yesNo=true);
public: // Different for each subclass
	unsigned int GetApproxMemUsage();
	bool SetVoteType(BuiltinVoteType voteType);
	void DisplayVotePass(const char *param1="");
	void DisplayVoteFail(BuiltinVoteFailReason voteReason);
	void CastVote(edict_t *pEntity, const CCommand &args);
protected: // Internal implementation of various actions inside the vote
	void ShowVote(cell_t clients[], unsigned int playerCount, const char *voteArgument);
	void ShowVoteCast(int entityId, int item);
	//void DisplayAllVotes(unsigned int totalClients, unsigned int yesVotes, unsigned int noVotes)
	void SendOptions();
};

extern IGameEventManager2 *events;
extern ICvar *icvar;

#endif //_INCLUDE_BUILTINVOTES_BASEVOTE_H_